#include <htc.h>
#include <xc.h>
#include <stdbool.h>

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover Mode (Internal/External Switchover Mode is disabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#define minutes 0x01
#define hours 0x02
#define ctrl 0x07

void initTmr(void);
void initPins(void);
void initI2C(void);
extern bool pressed(unsigned int number);
void updateLeds(void);
void updateModus(void);
void updateButtons(void);
void ADCRead(void);
void readTime(void);
void delay(void);
unsigned char read_ds1307(unsigned char address);
void write_ds1307(unsigned char address, unsigned char value);

typedef union
{
    unsigned char byte;

    struct
    {
        unsigned bit0 :1;
        unsigned bit1 :1;
        unsigned bit2 :1;
        unsigned bit3 :1;
        unsigned bit4 :1;
        unsigned bit5 :1;
        unsigned bit6 :1;
        unsigned bit7 :1;
    }bitwise;
}d2bcd;

typedef union
{
    unsigned int integer;

    struct
    {
        unsigned low  :8;
        unsigned high :8;
    }charwise;
}temp;

d2bcd ledDisplay1;
d2bcd ledDisplay2;
d2bcd ledDisplay3;
d2bcd ledDisplay4;

temp value;

unsigned char cntLed = 1;
unsigned char counter = 0;
unsigned char freq1 = 30;
unsigned char freq2 = 100;
unsigned char freq3 = 100;
unsigned char system = 24;
unsigned char readCounter = 0;
unsigned char tempCounter = 0;

unsigned char led1 = 0;
unsigned char led2 = 0;
unsigned char led3 = 0;
unsigned char led4 = 0;

unsigned char modus = 0;
unsigned char settings = 0;

unsigned int val = 0;

unsigned char temp1 = 0;
unsigned char temp2 = 0;

unsigned char ledHours = 0;
unsigned char ledMinutes = 0;

unsigned char readTemp = 0;

unsigned char changed = 0;

unsigned char res1 = 0;
unsigned char res2 = 0;

void main(void)
{
    initTmr();
    initPins();
    initI2C();

    readTime();

    while (1)
    {
        if (modus != 2)
        {
            updateLeds();
        }
        
        updateModus();
        updateButtons();
    }
}

void initTmr(void)
{
    OSCCONbits.IRCF = 0b1101;
    OSCCONbits.SCS = 0b10;

    OPTION_REGbits.TMR0CS = 0;
    OPTION_REGbits.PSA = 1;
    OPTION_REGbits.PS = 0b011;

    T1CONbits.TMR1CS = 0b00;
    T1CONbits.T1CKPS = 0b11;
    T1CONbits.TMR1ON = 1;

    T2CONbits.T2CKPS = 0b11;
    T2CONbits.T2OUTPS = 0b0000;
    T2CONbits.TMR2ON = 0;

    INTCONbits.TMR0IE = 1;
    PIE1bits.TMR1IE = 1;
    PIE1bits.TMR2IE = 1;

    PIR1bits.TMR1IF = 0;
    PIR1bits.TMR2IF = 0;
    INTCONbits.TMR0IF = 0;

    INTCONbits.GIE = 1;
    INTCONbits.PEIE = 1;
}

void initPins(void)
{
    ANSELA = 0x00;
    ANSELB = 0x00;
    ANSELC = 0x00;

    // Temp.-Sensor
    TRISCbits.TRISC6 = 1;
    ANSELCbits.ANSC6 = 1;

    // Buttons
    OPTION_REGbits.nWPUEN = 0;
    
    TRISAbits.TRISA2 = 1;
    TRISAbits.TRISA5 = 1;
    TRISAbits.TRISA4 = 1;
    TRISBbits.TRISB5 = 1;

    WPUAbits.WPUA2 = 1;
    WPUAbits.WPUA4 = 1;
    WPUAbits.WPUA5 = 1;
    WPUBbits.WPUB5 = 1;

    // Transistoren
    TRISCbits.TRISC0 = 0;
    TRISCbits.TRISC1 = 0;
    TRISCbits.TRISC2 = 0;
    TRISCbits.TRISC3 = 0;

    // BCD
    TRISCbits.TRISC7 = 0;
    TRISBbits.TRISB7 = 0;
    TRISCbits.TRISC5 = 0;
    TRISCbits.TRISC4 = 0;
}

void initI2C(void)
{
    TRISBbits.TRISB4 = 1;
    TRISBbits.TRISB6 = 1;

    SSPADD = 39;
    SSP1CON1bits.SSPM = 0b1000;
    SSP1CON1bits.SSPEN = 1;
}

void readTime(void)
{   
    unsigned char res1 = read_ds1307(minutes);
    unsigned char res2 = read_ds1307(hours);

    ledMinutes = (res1 % 16) + (((char) res1 / 16) * 10);
    ledHours = (res2 % 16) + (((char) res2 / 16) * 10);
}

void writeTime(void)
{
    unsigned char send1 = ((char) ledMinutes / 10) * 16 + ledMinutes % 10;
    unsigned char send2 = ((char) ledHours / 10) * 16 + ledHours % 10;
    
    write_ds1307(minutes, send1);
    write_ds1307(hours, send2);
}

void ADCRead(void)
{
   ADCON1bits.ADPREF = 0b00;
   ADCON0bits.ADON = 1;
   ADCON0bits.CHS = 0b1000;
   delay();

   ADCON1bits.ADCS = 0b010;
   ADCON1bits.ADFM = 1;

   delay();
   ADCON0bits.GO_nDONE = 1;
   while(ADCON0bits.GO_nDONE);

   delay();
   ADCON0bits.ADON = 0;


   char low = ADRESL;
   char high = ADRESH;

   value.charwise.high = high;
   value.charwise.low = low;
}

void updateButtons(void)
{
    if (pressed(1) && settings == 0)
        {
            modus++;
            if (modus > 2)
            {
                modus = 0;
            }
        }
        if (pressed(2) && modus == 0)
        {
            settings++;
            if (settings == 6)
            {
                settings = 0;
            }
        }
        if (pressed(3))
        {
            if (settings == 1)
            {
                ledHours++;
                if (ledHours == system)
                {
                    ledHours = 0;
                }
            }
            if (settings == 2)
            {
                ledMinutes++;
                if (ledMinutes == 60)
                {
                    ledMinutes = 0;
                }
            }
            if (settings == 3)
            {
                freq1++;
            }
            if (settings == 4)
            {
                freq2++;
            }
            if (settings == 5)
            {
                freq3++;
            }
        }
        if (pressed(4))
        {
            if (settings == 1)
            {
                if (ledHours == 0)
                {
                    ledHours = system;
                }
                ledHours--;
            }
            if (settings == 2)
            {
                if (ledMinutes == 0)
                {
                    ledMinutes = 60;
                }
                ledMinutes--;
            }
            if (settings == 3)
            {
                freq1--;
            }
            if (settings == 4)
            {
                freq2--;
            }
            if (settings == 5)
            {
                freq3--;
            }
        }
}

void updateLeds(void)
{
    switch (cntLed)
        {
            case 1:
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;

                ledDisplay1.byte = led1;

                PORTCbits.RC5 = ledDisplay1.bitwise.bit0;
                PORTBbits.RB7 = ledDisplay1.bitwise.bit1;
                PORTCbits.RC7 = ledDisplay1.bitwise.bit2;
                PORTCbits.RC4 = ledDisplay1.bitwise.bit3;

                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;
                PORTCbits.RC0 = 1;

                break;

            case 2:
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;

                ledDisplay2.byte = led2;

                PORTCbits.RC5 = ledDisplay2.bitwise.bit0;
                PORTBbits.RB7 = ledDisplay2.bitwise.bit1;
                PORTCbits.RC7 = ledDisplay2.bitwise.bit2;
                PORTCbits.RC4 = ledDisplay2.bitwise.bit3;

                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 1;

                break;

            case 3:
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;

                ledDisplay3.byte = led3;

                PORTCbits.RC5 = ledDisplay3.bitwise.bit0;
                PORTBbits.RB7 = ledDisplay3.bitwise.bit1;
                PORTCbits.RC7 = ledDisplay3.bitwise.bit2;
                PORTCbits.RC4 = ledDisplay3.bitwise.bit3;

                PORTCbits.RC3 = 0;
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 1;

                break;

            case 4:
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;
                
                ledDisplay4.byte = led4;

                PORTCbits.RC5 = ledDisplay4.bitwise.bit0;
                PORTBbits.RB7 = ledDisplay4.bitwise.bit1;
                PORTCbits.RC7 = ledDisplay4.bitwise.bit2;
                PORTCbits.RC4 = ledDisplay4.bitwise.bit3;

                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 1;

                break;

            default:
                PORTCbits.RC0 = 0;
                PORTCbits.RC1 = 0;
                PORTCbits.RC2 = 0;
                PORTCbits.RC3 = 0;

                break;
        }
}

void updateModus(void)
{
    if (modus == 0 && settings == 0)
    {
        led1 = (char) ledHours/10;
        led2 = (char) ledHours%10;
        led3 = (char) ledMinutes/10;
        led4 = (char) ledMinutes%10;

        INTCONbits.PEIE = 1;
        INTCONbits.GIE = 1;
    }

    if (modus == 1 && settings == 0)
    {
        readTemp = 1;
        T2CONbits.TMR2ON = 1;
        
        led1 = '.';
        led2 = '.';
        led3 = temp1;
        led4 = temp2;
    }

    if (modus == 2 && settings == 0)
    {
        T2CONbits.TMR2ON = 0;
        readTemp = 0;
        
        led1 = '.';
        led2 = '.';
        led3 = '.';
        led4 = '.';

        PORTCbits.RC0 = 0;
        PORTCbits.RC1 = 0;
        PORTCbits.RC2 = 0;
        PORTCbits.RC3 = 0;

        INTCONbits.PEIE = 0;
        INTCONbits.GIE = 0;
    }

    if (settings == 1)
    {
        led1 = (char) ledHours/10;
        led2 = (char) ledHours%10;
        led3 = '.';
        led4 = '.';
    }

    if (settings == 2)
    {
        led1 = '.';
        led2 = '.';
        led3 = (char) ledMinutes/10;
        led4 = (char) ledMinutes%10;

        changed = 1;
    }

    if (settings == 3)
    {
        if (changed == 1)
        {
            writeTime();
            changed = 0;
        }
        
        led1 = '.';
        led2 = (char) freq1/100;
        led3 = (char) (freq1 - (led2 * 100))/10;
        led4 = (char) freq1%10;
    }

    if (settings == 4)
    {
        led1 = '.';
        led2 = (char) freq2/100;
        led3 = (char) (freq2 - (led2 * 100))/10;
        led4 = (char) freq2%10;
    }

    if (settings == 5)
    {
        led1 = '.';
        led2 = (char) freq3/100;
        led3 = (char) (freq3 - (led2 * 100))/10;
        led4 = (char) freq3%10;
    }
}

void interrupt ISR (void)
{
    if (INTCONbits.TMR0IF == 1 && INTCONbits.TMR0IE == 1)
    {
        if (counter == freq1)
        {
            counter = 255;
            cntLed++;
            if (cntLed > 4)
            {
                cntLed = 1;
            }
        }
        counter++;
        
        INTCONbits.TMR0IF = 0;
    }

    if (PIE1bits.TMR1IE == 1 && PIR1bits.TMR1IF == 1)
    {
        if (readCounter == freq2)
        {
            readCounter == 255;
            readTime();
        }
        readCounter++;
        
        PIR1bits.TMR1IF = 0;
    }

    if (PIE1bits.TMR2IE == 1 && PIR1bits.TMR2IF == 1)
    {
        if (readTemp == 1)
        {
            if (tempCounter == freq3)
            {
                tempCounter = 255;
                ADCRead();
                val = value.integer;
                temp1 = (char) (val*0.4883)/10;
                temp2 = (char) (val*0.4883)%10;
            }
            tempCounter++;
        }
        
        PIR1bits.TMR2IF = 0;
    }
}

void delay(void)
{
    char i;
    for (i = 0; i < 255; i++)
    {
        // NOP
    }
    for (i = 0; i < 255; i++)
    {
        // NOP
    }
    for (i = 0; i < 255; i++)
    {
        // NOP
    }
}

unsigned char read_ds1307(unsigned char address)
{
    unsigned char data;
    SSP1CON2bits.SEN = 1;
    
    SSPBUF = 0xD0;
    while(SSP1CON2bits.ACKSTAT);

    SSPBUF = address;
    while(SSP1CON2bits.ACKSTAT);
    
    SSP1CON2bits.RSEN = 1;
    
    SSPBUF = 0xD1;
    while(SSP1CON2bits.ACKSTAT);

    SSP1CON2bits.RCEN = 1;

    data = SSPBUF;
    while (!BF);
    
    SSPCON2bits.ACKDT = 1;
    SSP1CON2bits.ACKEN = 1;
    SSP1CON2bits.PEN = 1;
 
    return(data);
}

void write_ds1307(unsigned char address, unsigned char value)
{
    SSP1CON2bits.SEN = 1;

    SSPBUF = 0xD1;
    while(SSP1CON2bits.ACKSTAT);

    SSPBUF = address;
    while(SSP1CON2bits.ACKSTAT);

    SSPBUF = value;
    while(SSP1CON2bits.ACKSTAT);
    
    SSP1CON2bits.PEN = 1;
}
